<?php

require_once 'parse_utils.php';
require_once 'mdb_types.php';
require_once 'dunemdb_api.php';

class DunemdbCache
{
    private static $instance = null;

    public static function get_instance()
    {
        if (!isset(self::$instance))
            self::$instance = new DunemdbCache();
        return self::$instance;
    }

    ///////////////////////////////////////////////////////////////////////

    const PUBS_DEBUG = false;

    const NUM_CACHED_FULL_OBJECTS = 10;
    const DEFAULT_GENRE_POSITION = 10000;

    ///////////////////////////////////////////////////////////////////////

    public static function get_base_dir()
    {
        return DataModel::get_storage_dir() . "/dunemdb_cache";
    }

    public static function get_movies_dir($ndx=0)
    {
        return $ndx == 1 && !DataModel::$with_persistfs ? null :
            DataModel::get_storage_dir($ndx) . "/dunemdb_cache/movies";
    }

    public static function get_genres_file($ndx=0)
    {
        return $ndx == 1 && !DataModel::$with_persistfs ? null :
            DataModel::get_storage_dir($ndx) . "/dunemdb_cache/genres.json";
    }

    public static function get_movie_posters_base_dir($ndx=0)
    {
        return $ndx == 1 && !DataModel::$with_persistfs ? null :
            DataModel::get_storage_dir($ndx) . "/image_cache/mdb/movie_poster";
    }

    public static function get_season_posters_base_dir($ndx=0)
    {
        return $ndx == 1 && !DataModel::$with_persistfs ? null :
            DataModel::get_storage_dir($ndx) . "/image_cache/mdb/season_poster";
    }

    public static function get_fanart_base_dir($ndx=0)
    {
        return $ndx == 1 && !DataModel::$with_persistfs ? null :
            DataModel::get_storage_dir($ndx) . "/image_cache/mdb/movie_fanart";
    }

    public static function get_person_photo_base_dir($ndx=0)
    {
        return $ndx == 1 && !DataModel::$with_persistfs ? null :
            DataModel::get_storage_dir($ndx) . "/image_cache/mdb/person_photo";
    }

    public static function get_episode_image_base_dir($ndx=0)
    {
        return $ndx == 1 && !DataModel::$with_persistfs ? null :
            DataModel::get_storage_dir($ndx) . "/image_cache/mdb/episode_image";
    }

    ///////////////////////////////////////////////////////////////////////

    private $genre_name_by_id;
    private $other_genre_id;
    private $unknown_genre_ids;
    private $genre_movie_count_by_id;
    private $mdb_genre_ids;

    public $movie_by_id;

    public $person_by_id;
    public $pmp_by_key;

    public $movie_ids_by_person_id;
    public $related_movie_ids_by_movie_id;

    private $last_played_movie_id;

    private $last_visited_ids;
    private $lvi_counter;

    private function __construct()
    {
        $dir = self::get_base_dir();
        if (!is_dir($dir))
        {
            if (mkdir($dir, 0777) === FALSE)
                hd_print("[dunemdb_cache] Error creating directory '$dir'");
        }

        $dir = self::get_movies_dir();
        if (!is_dir($dir))
        {
            if (mkdir($dir, 0777) === FALSE)
                hd_print("[dunemdb_cache] Error creating directory '$dir'");
        }

        $this->movie_by_id = array();

        $this->person_by_id = array();
        $this->pmp_by_key = array();

        $this->movie_ids_by_person_id = array();
        $this->related_movie_ids_by_movie_id = array();

        $this->last_visited_ids = array();
        $this->lvi_counter = 1;
    }

    public function set_object_visited($id)
    {
        $this->last_visited_ids[$id] = $this->lvi_counter;
        $this->lvi_counter++;

        if (count($this->last_visited_ids) > self::NUM_CACHED_FULL_OBJECTS)
        {
            asort($this->last_visited_ids, SORT_NUMERIC);
            list($key, $val) = each($this->last_visited_ids);
            hd_print("DEBUG: outdated visited id $key");
            unset($this->last_visited_ids[$key]);
        }
    }

    public function set_last_played_movie_id($movie_id)
    {
        $this->last_played_movie_id = $movie_id;
    }

    public function ensure_genres_loaded()
    {
        if (isset($genre_name_by_id))
            return;

        $this->genre_name_by_id = array();
        $this->unknown_genre_ids = array();
        $this->other_genre_id = null;
        $this->genre_movie_count_by_id = array();
        $this->mdb_genre_ids = array();

        $result = DunemdbApi::get_instance()->
            get_result_from_cache(
                self::get_genres_file(0), self::get_genres_file(1),
                DunemdbApi::GENRES_CACHE_EXPIRE_SEC);

        if (isset($result) && is_array($result))
        {
            foreach ($result as $item)
            {
                $id = strval($item->id);
                if ($id === '')
                    continue;

                $key = trim(strval($item->name));
                if (!isset(Genre::$VALUES[$key]))
                {
                    hd_print("Warning: unknown genre: $key");
                    $this->genre_name_by_id[$id] = $key;
                    $this->unknown_genre_ids[$id] = 1;
                }
                else
                {
                    $val = Genre::$VALUES[$key];
                    $this->genre_name_by_id[$id] = $val;
                    if ($val == Genre::$UNKNOWN_GENRE_VALUE)
                        $this->other_genre_id = $id;
                }

                $movieCount = intval($item->movieCount);
                $this->genre_movie_count_by_id[$id] = $movieCount;
            }
        }

        $this->mdb_genre_ids = array_keys($this->genre_name_by_id);
        usort($this->mdb_genre_ids, 'DataModel::cmp_genre_ids');
    }

    public function get_mdb_genre_ids()
    {
        return $this->mdb_genre_ids;
    }

    public function get_genre_name_by_id($genre_id, $with_translation = true)
    {
        if (!isset($this->genre_name_by_id[$genre_id]))
            return $genre_id;

        $genre_name = $this->genre_name_by_id[$genre_id];
        if ($with_translation && !isset($this->unknown_genre_ids[$genre_id]))
            $genre_name = "<key_local>$genre_name</key_local>";

        return $genre_name;
    }

    public function get_genre_position_by_id($genre_id)
    {
        $cur_genre_name = $this->get_genre_name_by_id($genre_id, false);
        foreach (Genre::$BY_POSITION as $i => $genre_name)
        {
            if ($cur_genre_name == $genre_name)
                return $i;
        }
        return self::DEFAULT_GENRE_POSITION;
    }

    public function is_other_genre_id($genre_id)
    {
        return isset($this->other_genre_id) &&
            $genre_id == $this->other_genre_id;
    }

    public function get_genre_movie_count_by_id($genre_id)
    {
        if (!isset($this->genre_movie_count_by_id[$genre_id]))
            return 0;
        return $this->genre_movie_count_by_id[$genre_id];
    }

    public function get_short_movie($dunemdb_id)
    {
        if (isset($this->movie_by_id[$dunemdb_id]))
            return $this->movie_by_id[$dunemdb_id];

        return $this->get_movie_from_disk_cache($dunemdb_id, false);
    }

    public function get_and_cache_short_movie($dunemdb_id)
    {
        if (isset($this->movie_by_id[$dunemdb_id]))
            return $this->movie_by_id[$dunemdb_id];

        $movie = $this->get_movie_from_disk_cache($dunemdb_id, false);
        if (!isset($movie))
            return null;

        $this->movie_by_id[$movie->id] = $movie;
        return $movie;
    }

    public function get_full_movie_with_connections($dunemdb_id,
        $use_disk_cache=true)
    {
        if (isset($this->movie_by_id[$dunemdb_id]))
        {
            $movie = $this->movie_by_id[$dunemdb_id];
            if (isset($movie->ext) && isset($movie->conn))
                return $movie;
        }

        $movie = $use_disk_cache ?
            $this->get_movie_from_disk_cache($dunemdb_id, true) : null;
        if ($movie != null)
        {
            $this->movie_by_id[$movie->id] = $movie;
            return $movie;
        }

        return null;
    }

    private static function parse_season_num($s)
    {
        if (!isset($s->number) || 0 == strlen($s->number))
        {
            if (!isset($s->year) || $s->year <= 0)
                return null;
            return "year:$s->year";
        }
        return intval($s->number);
    }

    private static function parse_season_poster_url($s, $lang_code)
    {
        return ParseUtils::parse_multilang_string($s, 'imgPoster', 
            $lang_code, 'missing://', 0, true);
    }

    private function do_parse_short_seasons($movie_id, $result)
    {
        $lang_code = SystemConfig::get_instance()->lang_code;

        $seasons = array();
        foreach ($result as $s)
        {
            $num = self::parse_season_num($s);
            if (!isset($num))
                continue;
            $poster_url = self::parse_season_poster_url($s, $lang_code);

            $year = isset($s->year) && $s->year > 0 ? $s->year : null;

            $episodes_count = isset($s->episodesCount) ?
                intval($s->episodesCount) : 0;
            if ($episodes_count <= 0)
            {
                hd_print("Warning: ignoring season $num without episodes");
                continue;
            }

            $seasons[$num] = new MdbMovieShortSeason($movie_id, $num,
                $poster_url, $episodes_count, $year);
        }
        uasort($seasons, 'mdb_cmp_seasons_by_number');
        return $seasons;
    }

    private function do_parse_seasons($movie_id, $result)
    {
        $lang_code = SystemConfig::get_instance()->lang_code;

        $seasons = array();
        foreach ($result as $s)
        {
            $snum = self::parse_season_num($s);
            if (!isset($snum))
                continue;
            $poster_url = self::parse_season_poster_url($s, $lang_code);

            $season = new MdbMovieSeason($movie_id, $snum, $poster_url);

            $season->year = null;
            foreach ($s->episodes as $e)
            {
                $ep_num = $e->number;
                $title = ParseUtils::parse_multilang_string($e, 'title', 
                    $lang_code, '', 1, true);
                $desc = ParseUtils::parse_multilang_string($e,
                    'description', $lang_code, '', 1, false);
                $premiered = isset($e->premiered) ? $e->premiered : '';
                $image_url = isset($e->imgBackdrops) &&
                    count($e->imgBackdrops) > 0 ?
                    $e->imgBackdrops[0] : 'missing://';

                if ($premiered && preg_match('|^(\d\d\d\d)-\d\d-\d\d$|', $premiered, $m))
                {
                    $y = $m[1];
                    if (!isset($season->year) || $season->year > $y)
                        $season->year = $y;
                }

                $season->episodes[$ep_num] = new MdbMovieEpisode(
                    $ep_num, $title, $desc, $image_url, $premiered);
            }
            $seasons[$snum] = $season;
        }
        uasort($seasons, 'mdb_cmp_seasons_by_number');
        return $seasons;
    }

    private function get_movie_from_disk_cache($dunemdb_id, $with_connections)
    {
        $path = self::get_movies_dir() . "/$dunemdb_id.json";
        if (!is_file($path))
        {
            if (DataModel::$with_persistfs)
            {
                $path = self::get_movies_dir(1) . "/$dunemdb_id.json";
                if (!is_file($path))
                    return null;
            }
            else
                return null;
        }

        $doc = file_get_contents($path);
        if ($doc === FALSE)
        {
            hd_print("[dunemdb_cache[ Failed to load $path.");
            return null;
        }

        $movie = $this->parse_movie_json($doc, $with_connections);
        if (isset($movie) && $with_connections)
            $movie->from_cache = true;
        return $movie;
    }

    public function update_movie_from_disk_cache($movie_id)
    {
        $movie = $this->get_movie_from_disk_cache($movie_id, true);
        if (isset($movie))
            $this->movie_by_id[$movie_id] = $movie;
        return $movie;
    }

    private function do_parse_movie_person(&$json, $related_movie_id)
    {
        $lang_code = SystemConfig::get_instance()->lang_code;

        $person_id = strval($json->id);

        if (isset($this->person_by_id[$person_id]))
            $person = $this->person_by_id[$person_id];
        else
        {
            $person = new MdbPerson(true, $person_id);
            $this->person_by_id[$person_id] = $person;
        }

        $person->name = ParseUtils::parse_multilang_string($json, 'name',
            $lang_code, '', 1, true);
        $person->motherland = isset($json->motherland) ?
            self::parse_country_value($json->motherland, true) : "";
        $person->birthday = isset($json->birthday) && $json->birthday != 0 ?
            intval($json->birthday) : HD::INT_MIN;
        $person->photo_url = ParseUtils::parse_multilang_string($json, 'imgPhoto',
            $lang_code, 'missing://', 0, true);

        if ($related_movie_id != null)
        {
            $key = "$related_movie_id.$person_id";
            if (!isset($this->pmp_by_key[$key]))
            {
                $pmp = new MdbPersonMovieParticipation();
                $this->pmp_by_key[$key] = $pmp;
            }

            $pmp = &$this->pmp_by_key[$key];

            if (isset($json->type))
            {
                if ($json->type == MdbPersonMovieJob::ACTOR)
                {
                    $pmp->role = ParseUtils::parse_multilang_string($json, 'role',
                        $lang_code, '', 1, false);
                }
                $job = intval($json->type);
                if ($job >= MdbPersonMovieJob::MIN_VALUE &&
                    $job <= MdbPersonMovieJob::MAX_VALUE)
                {
                    if (!in_array($job, $pmp->jobs))
                        $pmp->jobs[] = $job;
                }
            }
        }

        return $person;
    }

    private function do_fill_people_string(&$out_str,
        &$people, $required_type, $max_len, $lang_code)
    {
        $total_len = 0;
        foreach ($people as $person)
        {
            $type = isset($person->type) ? intval($person->type) : -1;
            if ($type !== $required_type)
                continue;

            $name = ParseUtils::parse_multilang_string(
                $person, 'name', $lang_code, '', 1, true);

            $new_len = mb_strlen($name, 'UTF-8') +
                ($total_len == 0 ? 0 : 2);
            if ($total_len + $new_len > $max_len)
                break;

            if ($total_len > 0)
                $out_str .= ", ";
            $out_str .= $name;

            $total_len += $new_len;
        }
    }

    private static function parse_country_value($value, $single_value)
    {
        $country = strtolower(strval($value));
        if ($country === '')
            return '';

        if (isset(Country::$VALUES[$country]))
        {
            return ($single_value ? "%ext%" : "").
                "<key_local>".
                Country::$VALUES[$country].
                "</key_local>";
        }

        if ($country[0] == '_')
            return substr($country, 1);
        return $country;
    }

    private function do_parse_movie_json($partial,
        $with_connections, $related_person_id, &$result)
    {
        $lang_code = SystemConfig::get_instance()->lang_code;

        $countries_str = "";
        $genres_str = "";
        $desc_lang_code = '';
        $alt_desc_lang_code = '';
        $desc = '';

        $alt_desc = '';

        $kp_id = '';
        $imdb_id = '';
        $tmdb_id = '';
        $tmdbtv_id = '';
        if (isset($result->sources))
        {
            if (isset($result->sources->Kinopoisk))
                $kp_id = strval($result->sources->Kinopoisk);
            if (isset($result->sources->IMDB))
                $imdb_id = strval($result->sources->IMDB);
            if (isset($result->sources->TheMovieDB))
                $tmdb_id = strval($result->sources->TheMovieDB);
            if (isset($result->sources->TheMovieDBTV))
                $tmdbtv_id = strval($result->sources->TheMovieDBTV);
        }
        if (isset($result->auxSources))
        {
            if ($kp_id == '' && isset($result->auxSources->Kinopoisk))
                $kp_id = strval($result->auxSources->Kinopoisk);
            if ($imdb_id == '' && isset($result->auxSources->IMDB))
                $imdb_id = strval($result->auxSources->IMDB);
            if ($tmdb_id == '' && isset($result->auxSources->TheMovieDB))
                $tmdb_id = strval($result->auxSources->TheMovieDB);
            if ($tmdbtv_id == '' && isset($result->auxSources->TheMovieDBTV))
                $tmdbtv_id = strval($result->auxSources->TheMovieDBTV);
        }

        $poster_lang = null;
        $poster_url = 'missing://';

        $fanart_url = 'missing://';

        $movie_id = strval($result->id);

        $title_lang = null;
        $title = ParseUtils::parse_multilang_string($result, 'title', 
            $lang_code, '', 1, true, $title_lang);

        $series_type = isset($result->type) &&
            strval($result->type) == 'series';

        $native_title = '';

        $movie = new MdbMovie($movie_id,
            $title, $series_type, $native_title,
            $desc, $alt_desc,
            $poster_url, $poster_lang,
            'missing://',
            $fanart_url,
            isset($result->year) && $result->year > 0 ?
                strval($result->year) : "",
            isset($result->endYear) && $result->endYear > 0 ?
                strval($result->endYear) : "",
            isset($result->rating->imdb->rating) ?
                strval($result->rating->imdb->rating) :
                (isset($result->rating->imdb) && is_string($result->rating->imdb) ?
                 strval($result->rating->imdb) : ""),
            isset($result->rating->kp->rating) ?
                strval($result->rating->kp->rating) :
                (isset($result->rating->kp) && is_string($result->rating->kp) ?
                 strval($result->rating->kp) : ""),
            "",
            $countries_str,
            isset($result->duration) ?
                strval($result->duration) : "0",
            $genres_str,
            false,
            $kp_id, $imdb_id, $tmdb_id, $tmdbtv_id
        );

        return $movie;
    }

    private function parse_movie_json($txt, $with_connections)
    {
        $response = json_decode($txt);
        if (!isset($response) || !isset($response->result))
        {
            hd_print("[dunemdb_cache] Failed to parse movie json:\n".
                substr($txt, 0, 1000) . (strlen($txt) > 1000 ? "..." : ""));
            return null;
        }

        return $this->do_parse_movie_json(
            false, $with_connections, null, $response->result);
    }

    ///////////////////////////////////////////////////////////////////////

    public function parse_short_movie_list(&$json)
    {
        foreach ($json as $item)
        {
            $movie = $this->do_parse_movie_json(
                true, false, null, $item);
            if (!isset($movie))
                continue;

            $movie_ids[] = $movie->id;

            if (!isset($this->movie_by_id[$movie->id]))
                $this->movie_by_id[$movie->id] = $movie;
        }

        return isset($movie_ids) ? $movie_ids : null;
    }

    public function parse_short_movie(&$json)
    {
        $movie = $this->do_parse_movie_json(
            true, false, null, $json);
        if (!isset($movie))
            return null;

        if (!isset($this->movie_by_id[$movie->id]))
            $this->movie_by_id[$movie->id] = $movie;

        return $movie->id;
    }

    ///////////////////////////////////////////////////////////////////////

    public function full_movie_is_cached($movie_id)
    {
        $cache_path = self::get_movies_dir() . "/$movie_id.json";
        if (is_file($cache_path))
            return true;

        if (DataModel::$with_persistfs)
        {
            $cache_path = self::get_movies_dir(1) . "/$movie_id.json";
            return is_file($cache_path);
        }
        else
            return false;
    }

    public function get_cached_movie_genre_ids($movie_id)
    {
        $cache_path = self::get_movies_dir() . "/$movie_id.json";
        $cache_path2 = DataModel::$with_persistfs ?
            self::get_movies_dir(1) . "/$movie_id.json" : null;
        $result = DunemdbApi::get_instance()->
            get_result_from_cache($cache_path, $cache_path2);
        if (!isset($result))
            return null;
        if (!isset($result->genres))
            return null;

        $genre_ids = array();
        foreach ($result->genres as $genre)
        {
            $genre_ids[] = isset($genre->id) ?
                strval($genre->id) : strval($genre);
        }
        return $genre_ids;
    }
}

?>
